@@ -1,14 +1,17 @@
 """This module is capable of creating and signing Exonum transactions."""
 
-from typing import Dict, Optional
+from typing import Dict, Optional, Tuple, Any
 import json
-import struct
+from logging import getLogger
 
 from google.protobuf.message import Message as ProtobufMessage, DecodeError as ProtobufDecodeError
 
 from .crypto import PublicKey, Hash, Signature, KeyPair
 from .module_manager import ModuleManager
 
+# pylint: disable=C0103
+logger = getLogger(__name__)
+
 
 class MessageGenerator:
     """MessageGenerator is a class which helps you create transactions.
@@ -18,18 +21,19 @@ class MessageGenerator:
     Example of usage:
     >>> instance_id = ... # Get the ID of the desired service instance.
     >>> artifact_name = ... # Get the name of the artifact (not the instance).
-    >>> cryptocurrency_message_generator = MessageGenerator(instance_id, artifact_name) # Create a message generator.
-    >>> create_wallet_alice = cryptocurrency_module.CreateWallet() # Create a Protobuf message.
+    >>> artifact_version = ... # Get the version of the artifact (not the instance).
+    >>> cryptocurrency_message_generator = MessageGenerator(instance_id, artifact_name, artifact_version)
+    >>> create_wallet_alice = cryptocurrency_message_generator.CreateWallet() # Create a Protobuf message.
     >>> create_wallet_alice.name = "Alice1" # Fill the Protobuf message manually.
 
     Then you can transform the Protobuf message into an Exonum transaction.
 
     >>> create_wallet_alice_tx = cryptocurrency_message_generator.create_message(create_wallet_alice)
     >>> create_wallet_alice_tx.sign(keypair) # You should sign the message before sending.
-    >>> client.send_transaction(create_wallet_alice_tx)
+    >>> client.public_api.send_transaction(create_wallet_alice_tx)
     """
 
-    def __init__(self, instance_id: int, artifact_name: str, module_name: str):
+    def __init__(self, instance_id: int, artifact_name: str, artifact_version: str):
         """MessageGenerator constructor.
 
         Parameters
@@ -37,16 +41,15 @@ class MessageGenerator:
         instance_id: int
             ID of the desired Exonum service instance.
         artifact_name: str
-            The name of the service artifact you want to communicate with (e.g. 'cryptocurrency-advanced').
-        module_name: str
-            The name of the .proto file (e.g. "cryptocurrency") to load messages from.
+            The name of the service artifact you want to communicate with.
+            The name should be in the format provided by Exonum, like 'exonum-cryptocurrency-advanced:1.0.0'.
         """
         self._instance_id = instance_id
 
         self._artifact_name = artifact_name
         self._message_ids: Dict[str, int] = dict()
 
-        service_module = ModuleManager.import_service_module(artifact_name, module_name)
+        service_module = ModuleManager.import_service_module(artifact_name, artifact_version, "service")
         for i, message in enumerate(service_module.DESCRIPTOR.message_types_by_name):
             self._message_ids[message] = i
 
@@ -67,6 +70,18 @@ class MessageGenerator:
         tx_name = type(message).__name__
         return ExonumMessage(self._instance_id, self._message_ids[tx_name], message)
 
+    @staticmethod
+    def pk_to_hash_address(public_key: PublicKey) -> Optional[Hash]:
+        """Converts `PublicKey` into a `Hash`, which is a uniform
+        presentation of any transaction authorization supported by Exonum."""
+        types_module = ModuleManager.import_main_module("exonum.crypto.types")
+        auth_module = ModuleManager.import_main_module("exonum.runtime.auth")
+
+        caller = auth_module.Caller()
+        caller.transaction_author.CopyFrom(types_module.PublicKey(data=public_key.value))
+        hash_address = Hash.hash_data(caller.SerializeToString())
+        return hash_address
+
 
 class ExonumMessage:
     """Generic Exonum transaction class.
@@ -86,7 +101,7 @@ class ExonumMessage:
     >>> create_wallet_alice.name = "Alice1" # Fill the Protobuf message manually.
     >>> create_wallet_alice_tx = cryptocurrency_message_generator.create_message(create_wallet_alice)
     >>> create_wallet_alice_tx.sign(keypair) # You should sign the message before sending.
-    >>> client.send_transaction(create_wallet_alice_tx)
+    >>> client.public_api.send_transaction(create_wallet_alice_tx)
 
     Parsing a message:
 
@@ -106,14 +121,6 @@ class ExonumMessage:
     >>> tx_json = message.pack_into_json() # Create a JSON with the transaction in the format expected by Exonum.
     """
 
-    _MINIMUM_TX_BODY_LENGTH = 102  # It calculated as first 38 metadata bytes plus signature with 64 bytes length
-    _PUBLIC_KEY_LENGTH = 32
-    _SIGNATURE_LENGTH = 64
-    _SERVICE_ID_START_POSITION = 34
-    _MESSAGE_ID_START_POSITION = 36
-    _PROTO_MESSAGE_START_POSITION = 38
-    _U16_LENGTH = 2
-
     def __init__(self, instance_id: int, message_id: int, msg: ProtobufMessage, prebuilt: Optional[bytes] = None):
         """Exonum message constructor. It is not intended to be used directly, see `MessageGenerator.create_message`
         and `ExonumMessage.from_hex` instead."""
@@ -133,24 +140,25 @@ class ExonumMessage:
 
         # If we parse the received message, we do not have to build anything.
         if prebuilt is None:
-            self._unsigned_tx_raw = self._build_message()
+            self._any_tx_raw = self._build_message()
         else:
-            self._unsigned_tx_raw = prebuilt
+            self._any_tx_raw = prebuilt
 
     @classmethod
     def from_hex(
-        cls, message_hex: str, artifact_name: str, module_name: str, tx_name: str
+        cls, message_hex: str, artifact_name: str, artifact_version: str, tx_name: str
     ) -> Optional["ExonumMessage"]:
-        """Attempts to parse Exonum Message from serialized hexadecimal string.
+        """Attempts to parse an Exonum message from a serialized hexadecimal string.
 
         Parameters
         ----------
         message_hex: str
             Serialized message as a hexadecimal string.
         artifact_name: str
-            The name of the service artifact you want to communicate with (e.g. 'cryptocurrency-advanced').
-        module_name: str
-            The name of the .proto file (e.g. "cryptocurrency") to load messages from.
+            The name of the service artifact you want to communicate with,
+            e.g. 'exonum-cryptocurrency-advanced'.
+        artifact_version: str
+            Version of artifact as string, e.g. '1.0.0'.
         tx_name: str
             The name of the transaction to be parsed, e.g. 'CreateWallet'.
 
@@ -160,38 +168,33 @@ class ExonumMessage:
             If parsing is successfull, an ExonumMessage object is returned.
             Otherwise the returned value is None.
         """
-        # Load modules and prepare expected message class for parsing.
-        service_mod = ModuleManager.import_service_module(artifact_name, module_name)
-        transaction_class = getattr(service_mod, tx_name)
-
-        # Convert message from hex to bytes.
-        tx_raw = bytes.fromhex(message_hex)
-
-        # Parse data.
-        if len(tx_raw) < cls._MINIMUM_TX_BODY_LENGTH:
-            return None
         try:
-            author = tx_raw[: cls._PUBLIC_KEY_LENGTH]
-            service_id = struct.unpack(
-                "<H", tx_raw[cls._SERVICE_ID_START_POSITION : cls._SERVICE_ID_START_POSITION + cls._U16_LENGTH]
-            )[0]
-            message_id = struct.unpack(
-                "<H", tx_raw[cls._MESSAGE_ID_START_POSITION : cls._MESSAGE_ID_START_POSITION + cls._U16_LENGTH]
-            )[0]
-            signature = tx_raw[-cls._SIGNATURE_LENGTH :]
+            signed_msg, exonum_msg, decoded_msg = cls._deserialize_message(
+                message_hex, artifact_name, artifact_version, tx_name
+            )
 
-            decoded_msg = transaction_class()
-            decoded_msg.ParseFromString(tx_raw[cls._PROTO_MESSAGE_START_POSITION : -cls._SIGNATURE_LENGTH])
-        except (ValueError, IndexError, ProtobufDecodeError):
-            return None
+            service_id = exonum_msg.any_tx.call_info.instance_id
+            message_id = exonum_msg.any_tx.call_info.method_id
+            signature = signed_msg.signature.data[:]
+            author = signed_msg.author.data[:]
 
-        # Create the message.
-        try:
-            exonum_message = cls(service_id, message_id, decoded_msg)
-            cls._set_signature_data(exonum_message, author, signature, tx_raw)
+            exonum_message = cls(service_id, message_id, decoded_msg, prebuilt=exonum_msg.any_tx.SerializeToString())
 
+            cls._set_signature_data(exonum_message, author, signature, bytes.fromhex(message_hex))
+            logger.debug(
+                "Exonum message (ID: %s) from the service artifact '%s' " "for transaction '%s' parsed successfully.",
+                message_id,
+                artifact_name,
+                tx_name,
+            )
             return exonum_message
-        except ValueError:
+        except ProtobufDecodeError as e:
+            logger.error(
+                "Failed to parse an Exonum message from the service artifact '%s' " "for transaction '%s'. Error: %s.",
+                artifact_name,
+                tx_name,
+                str(e),
+            )
             return None
 
     def sign(self, keys: KeyPair) -> None:
@@ -208,22 +211,27 @@ class ExonumMessage:
         public_key, secret_key = keys.public_key, keys.secret_key
         self._author = public_key
 
-        signed_raw = bytearray()
+        messages_mod = ModuleManager.import_main_module("exonum.messages")
+        types_mod = ModuleManager.import_main_module("exonum.crypto.types")
 
-        signed_raw.extend(public_key.value)
-        signed_raw.extend(struct.pack("<B", 0))  # 0 and 0 it's tag and class of TX message
-        signed_raw.extend(struct.pack("<B", 0))
-        signed_raw.extend(struct.pack("<H", self._instance_id))
-        signed_raw.extend(struct.pack("<H", self._message_id))
-        signed_raw.extend(self._unsigned_tx_raw)
+        signed_message = messages_mod.SignedMessage()
+        signed_message.payload = self._any_tx_raw
+        signed_message.author.CopyFrom(types_mod.PublicKey(data=public_key.value))
 
-        signature = Signature.sign(bytes(signed_raw), secret_key)
+        signature = Signature.sign(signed_message.payload, secret_key)
 
-        signed_raw.extend(signature.value)
+        signed_message.signature.CopyFrom(types_mod.Signature(data=signature.value))
 
         self._signature = signature
 
-        self._signed_tx_raw = bytes(signed_raw)
+        self._signed_tx_raw = bytes(signed_message.SerializeToString())
+
+        logger.debug(
+            "Successfully signed the message (message ID: %s, service instance ID: %s): public_key='%s'.",
+            self._message_id,
+            self._instance_id,
+            public_key,
+        )
 
     def validate(self) -> bool:
         """
@@ -231,12 +239,24 @@ class ExonumMessage:
         Checks if the transaction signature is correct.
         :return: bool
         """
-        if self._signature is None or self._author is None or self._signed_tx_raw is None:
+        if self._signature is None or self._author is None:
             return False
 
         try:
-            return self._signature.verify(self._signed_tx_raw[: -self._SIGNATURE_LENGTH], self._author)
-        except (ProtobufDecodeError, ValueError, IndexError):
+            messages_mod = ModuleManager.import_main_module("exonum.messages")
+
+            signed_msg = messages_mod.SignedMessage()
+            signed_msg.ParseFromString(self._signed_tx_raw)
+
+            return self._signature.verify(signed_msg.payload, self._author)
+        except (ProtobufDecodeError, ValueError) as e:
+            logger.error(
+                "Failed to parse Exonum message (message ID: %s, service instance ID: %s): public_key='%s'. Error: %s",
+                self._message_id,
+                self._instance_id,
+                self._author,
+                str(e),
+            )
             return False
 
     def pack_into_json(self) -> str:
@@ -255,7 +275,8 @@ class ExonumMessage:
             An error will be raised on attempt to call `pack_into_json` with an unsigned message.
         """
         if self._signed_tx_raw is None:
-            raise RuntimeError("Attempt to call `to_json` on an unsigned message.")
+            logger.critical("Attempt to call `to_json` on an unsigned message into JSON format.")
+            raise RuntimeError("Attempt to pack an unsigned message.")
         return json.dumps({"tx_body": self._signed_tx_raw.hex()}, indent=4)
 
     def hash(self) -> Hash:
@@ -272,9 +293,9 @@ class ExonumMessage:
         """Returns a signature. If the message is not signed, returns None."""
         return self._signature
 
-    def unsigned_raw(self) -> bytes:
-        """Returns a serialized unsigned message as bytes."""
-        return self._unsigned_tx_raw
+    def any_tx_raw(self) -> bytes:
+        """Returns a serialized AnyTx message as bytes."""
+        return self._any_tx_raw
 
     def signed_raw(self) -> Optional[bytes]:
         """Returns a serialized SignedMessage as bytes. If the message is not signed, returns None."""
@@ -286,5 +307,50 @@ class ExonumMessage:
         self._signed_tx_raw = raw
 
     def _build_message(self) -> bytes:
-        """Builds a raw (unsigned) message."""
-        return self._msg.SerializeToString()
+        """Builds a raw AnyTx message."""
+        base_mod = ModuleManager.import_main_module("exonum.runtime.base")
+        messages_mod = ModuleManager.import_main_module("exonum.messages")
+
+        serialized_msg = self._msg.SerializeToString()
+
+        call_info = base_mod.CallInfo()
+        call_info.instance_id = self._instance_id
+        call_info.method_id = self._message_id
+
+        any_tx = base_mod.AnyTx()
+        any_tx.call_info.CopyFrom(call_info)
+        any_tx.arguments = serialized_msg
+
+        exonum_message = messages_mod.CoreMessage()
+        exonum_message.any_tx.CopyFrom(any_tx)
+
+        return exonum_message.SerializeToString()
+
+    @staticmethod
+    def _deserialize_message(
+        message_hex: str, artifact_name: str, artifact_version: str, tx_name: str
+    ) -> Tuple[Any, Any, Any]:
+        """Takes a serialized message as an argument and returns a tuple
+        [SignedMessage, ExonumMessage,DecodedMessage]."""
+
+        # Load modules and prepare an expected message class for parsing.
+        messages_mod = ModuleManager.import_main_module("exonum.messages")
+        service_mod = ModuleManager.import_service_module(artifact_name, artifact_version, "service")
+        transaction_class = getattr(service_mod, tx_name)
+
+        # Convert a message from hex to bytes.
+        tx_raw = bytes.fromhex(message_hex)
+
+        # Parse SignedMessage.
+        signed_msg = messages_mod.SignedMessage()
+        signed_msg.ParseFromString(tx_raw)
+
+        # Parse ExonumMessage from the SignedMessage's payload.
+        exonum_msg = messages_mod.CoreMessage()
+        exonum_msg.ParseFromString(signed_msg.payload)
+
+        # Parse an expected message from ExonumMessage's AnyTx arguments.
+        decoded_msg = transaction_class()
+        decoded_msg.ParseFromString(exonum_msg.any_tx.arguments)
+
+        return signed_msg, exonum_msg, decoded_msg
