@@ -1,11 +1,15 @@
 """MapProofBuilder module."""
 from typing import Optional, Dict, Any
+from logging import getLogger
 
 from exonum_client.module_manager import ModuleManager
 from .map_proof import MapProof
 from .errors import MapProofBuilderError
 from ..encoder import build_encoder_function
 
+# pylint: disable=C0103
+logger = getLogger(__name__)
+
 
 class MapProofBuilder:
     """
@@ -18,10 +22,14 @@ class MapProofBuilder:
     Designed workflow example:
 
     >>> proof_builder = MapProofBuilder()
-    >>> proof_builder.set_key_encoder('BitVec', main_module='helpers')
+    >>> proof_builder.set_key_encoder('BitVec', main_module='types')
     >>> proof_builder.set_value_encoder('Wallet', service_name='cryptocurrency', service_module='service')
     >>> proof = proof_builder.build_proof(map_proof)
 
+    If the proof is built from `RawProofMapIndex`, you can provide bool `raw` variable to the constructor:
+
+    >>> proof_builder = MapProofBuilder(raw=True)
+
     If your key/value type does not require convertion to bytes through Protobuf, use `MapProof.parse`
     instead and provide your converter functions by yourself.
 
@@ -30,40 +38,53 @@ class MapProofBuilder:
     returns a converter function.
     """
 
-    def __init__(self) -> None:
+    def __init__(self, raw: bool = False) -> None:
         """ Constructor of MapProofBuilder. """
         self._key_encoder: Optional[type] = None
         self._value_encoder: Optional[type] = None
+        self._raw = raw
 
     @staticmethod
     def _get_encoder(
         structure_name: str,
         main_module: Optional[str] = None,
         service_name: Optional[str] = None,
+        service_version: Optional[str] = None,
         service_module: Optional[str] = None,
     ) -> type:
         try:
             if main_module:
                 module = ModuleManager.import_main_module(main_module)
-            elif service_name and service_module:
-                module = ModuleManager.import_service_module(service_name, service_module)
+            elif service_name and service_module and service_version:
+                module = ModuleManager.import_service_module(service_name, service_version, service_module)
             else:
-                raise MapProofBuilderError("Module data not provided")
+                err = MapProofBuilderError("Module data not provided")
+                logger.warning(str(err))
+                raise err
 
-            return getattr(module, structure_name)
+            encoder = getattr(module, structure_name)
+            logger.debug("Successfully got encoder.")
+            return encoder
         except (ModuleNotFoundError, ImportError):
             error_data = {"main_module": main_module, "service_name": service_name, "service_module": service_module}
 
-            raise MapProofBuilderError("Incorrect module data", error_data)
+            err = MapProofBuilderError("Incorrect module data", error_data)
+            logger.warning("%s: %s", str(err), error_data)
+            raise err
         except AttributeError:
             error_data = {"service_name": structure_name}
-            raise MapProofBuilderError("Incorrect structure name", error_data)
 
+            err = MapProofBuilderError("Incorrect structure name", error_data)
+            logger.warning("%s: %s", str(err), error_data)
+            raise err
+
+    # pylint: disable=too-many-arguments
     def set_key_encoder(
         self,
         structure_name: str,
         main_module: Optional[str] = None,
         service_name: Optional[str] = None,
+        service_version: Optional[str] = None,
         service_module: Optional[str] = None,
     ) -> "MapProofBuilder":
         """
@@ -81,6 +102,8 @@ class MapProofBuilder:
             Name of the main module.
         service_name: Optional[str]
             Name of the service.
+        service_version: Optional[str]
+            Version of the service.
         service_module: Optional[str]
             Name of the service module.
 
@@ -89,14 +112,18 @@ class MapProofBuilder:
         MapProofBuilderError
             If the provided data is incorrect, this exception rises.
         """
-        self._key_encoder = self._get_encoder(structure_name, main_module, service_name, service_module)
+        self._key_encoder = self._get_encoder(
+            structure_name, main_module, service_name, service_version, service_module
+        )
         return self
 
+    # pylint: disable=too-many-arguments
     def set_value_encoder(
         self,
         structure_name: str,
         main_module: Optional[str] = None,
         service_name: Optional[str] = None,
+        service_version: Optional[str] = None,
         service_module: Optional[str] = None,
     ) -> "MapProofBuilder":
         """
@@ -114,6 +141,8 @@ class MapProofBuilder:
             Name of the main module.
         service_name: Optional[str]
             Name of the service.
+        service_version: Optional[str]
+            Version of the service.
         service_module: Optional[str]
             Name of the service module.
 
@@ -122,7 +151,9 @@ class MapProofBuilder:
         MapProofBuilderError
             If the provided data is incorrect, this exception rises.
         """
-        self._value_encoder = self._get_encoder(structure_name, main_module, service_name, service_module)
+        self._value_encoder = self._get_encoder(
+            structure_name, main_module, service_name, service_version, service_module
+        )
 
         return self
 
@@ -148,9 +179,11 @@ class MapProofBuilder:
             rises.
         """
         if not self._key_encoder or not self._value_encoder:
-            raise MapProofBuilderError("Encoders are not set")
+            err = MapProofBuilderError("Encoders are not set.")
+            logger.warning(str(err))
+            raise err
 
         key_encoder_func = build_encoder_function(self._key_encoder)
         value_encoder_func = build_encoder_function(self._value_encoder)
 
-        return MapProof.parse(proof, key_encoder_func, value_encoder_func)
+        return MapProof.parse(proof, key_encoder_func, value_encoder_func, self._raw)
