@@ -1,8 +1,12 @@
 """Module with OptionalEntry of MapProof."""
 from typing import Optional, Any, Dict
+from logging import getLogger
 
 from .errors import MalformedMapProofError
 
+# pylint: disable=C0103
+logger = getLogger(__name__)
+
 
 class OptionalEntry:
     """Optional entry is an entry of MapProof which can either miss a key or a key/value pair."""
@@ -27,4 +31,5 @@ class OptionalEntry:
         if data.get("key") and data.get("value"):
             return OptionalEntry(key=data["key"], value=data["value"])
 
+        logger.warning("Failed to parse an OptionalEntry from the provided JSON dict.")
         raise MalformedMapProofError.malformed_entry(data)
