@@ -2,6 +2,7 @@
 
 ProtobufLoader is capable of downloading Protobuf sources from Exonum."""
 from typing import List, Optional, Any, NamedTuple
+from logging import getLogger
 import shutil
 import sys
 import os
@@ -10,6 +11,11 @@ import re
 
 from .protoc import Protoc
 
+# pylint: disable=C0103
+logger = getLogger(__name__)
+
+PYTHON_RUNTIME = 2
+
 
 class ProtoFile(NamedTuple):
     """Structure that represents a proto file."""
@@ -25,7 +31,9 @@ class ProtobufProviderInterface:
         """Gets the Exonum core proto sources."""
         raise NotImplementedError
 
-    def get_proto_sources_for_artifact(self, runtime_id: int, artifact_name: str) -> List[ProtoFile]:
+    def get_proto_sources_for_artifact(
+        self, runtime_id: int, artifact_name: str, artifact_version: str
+    ) -> List[ProtoFile]:
         """Gets the Exonum core proto sources."""
         raise NotImplementedError
 
@@ -39,7 +47,7 @@ class ProtobufLoader:
 
     >>> with client.protobuf_loader() as loader:
     >>>    loader.load_main_proto_files()
-    >>>    loader.load_service_proto_files(0, "service_name")
+    >>>    loader.load_service_proto_files(0, "exonum-supervisor", "1.0.0")
 
     Code above will initialize loader, download core Exonum proto files and proto files for the Supervisor service.
     The code will compile the files into the Python modules. After that you will be able to load those modules
@@ -53,7 +61,7 @@ class ProtobufLoader:
     >>> ... # Some code
     >>> loader.deinitialize()
 
-    If you forget to call `deinitialize` (or if the code enters early, for example because of unhandled exception),
+    If you forget to call `deinitialize` (or if the code exits earlier, for example because of unhandled exception),
     the recourses created in the temp folder (which may differ depending on your OS) will not be removed.
 
     Creating more than one entity at a time will result in retrieving the same object:
@@ -73,11 +81,11 @@ class ProtobufLoader:
 
     >>> # module_a.py
     >>> loader = ProtobufLoader() # Since loader is already initialized with the client, you do not have to provide it.
-    >>> loader.load_service_proto_files(runtime_a, service_a)
+    >>> loader.load_service_proto_files(runtime_a, service_a, service_a_version)
 
     >>> # module_b.py
     >>> loader = ProtobufLoader()
-    >>> loader.load_service_proto_files(runtime_b, service_b)
+    >>> loader.load_service_proto_files(runtime_b, service_b, service_b_version)
 
     However, if you try to create the second loader, different from the first one, from the client,
     ValueError will be raised.
@@ -98,11 +106,19 @@ class ProtobufLoader:
         # Check that the client is the same as expected:
         if ProtobufLoader._entity is not None:
             if client is not None and client != ProtobufLoader._entity.client:
-                raise ValueError("Attempt to create ProtobufLoader entity with a different client")
+                err_msg = (
+                    f"Attempt to create ProtobufLoader entity with a different client:\n"
+                    f"used client:\n{ProtobufLoader._entity.client}\n"
+                    f"provided client:\n{client}\n"
+                )
+                logger.critical(err_msg)
+                raise ValueError(err_msg)
             return
 
         if client is None:
-            raise ValueError("Client is expected to be set for the initial object creation")
+            err_msg = "Client is expected to be set for the initial object creation."
+            logger.critical(err_msg)
+            raise ValueError(err_msg)
 
         ProtobufLoader._entity = self
         self.client = client
@@ -122,6 +138,7 @@ class ProtobufLoader:
 
         # Update the reference counter:
         ProtobufLoader._reference_count += 1
+        logger.debug("Current ProtobufLoader reference count: %s.", ProtobufLoader._reference_count)
         if ProtobufLoader._reference_count > 1:
             # If this is a second (third, etc) entity, everything is already initialized:
             return
@@ -140,15 +157,20 @@ class ProtobufLoader:
         # Add a directory with exonum_modules into the Python path:
         sys.path.append(self._proto_dir)
 
+        logger.debug("Successfully initialized ProtobufLoader for client:\n%s\n", self.client)
+
     def deinitialize(self) -> None:
         """Performs a deinitialization process."""
         if self._proto_dir is None:
-            raise RuntimeError("Attempt to deinitialize unititialized ProtobufLoader")
+            err_msg = "Attempt to deinitialize uninitialized ProtobufLoader."
+            logger.critical(err_msg)
+            raise RuntimeError(err_msg)
 
         # Decrement the reference counter:
         ProtobufLoader._reference_count -= 1
+        logger.debug("Current ProtobufLoader reference count: %s.", ProtobufLoader._reference_count)
 
-        # If there is still more than one reference, nothing should be done:
+        # If there is at least one reference, nothing should be done:
         if ProtobufLoader._reference_count > 0:
             return
 
@@ -159,21 +181,31 @@ class ProtobufLoader:
         sys.path.remove(self._proto_dir)
         shutil.rmtree(self._proto_dir)
 
+        # Unload any previously loaded protobuf modules
+        loaded_modules = list(sys.modules.keys())
+        for module in loaded_modules:
+            if module.startswith("exonum_modules"):
+                del sys.modules[module]
+
+        logger.debug("Successfully deinitialized ProtobufLoader.")
+
     @staticmethod
     def _save_proto_file(path: str, file_content: str) -> None:
         with open(path, "wt") as file_out:
             file_out.write(file_content)
 
     def _save_files(self, path: str, files: List[ProtoFile]) -> None:
-        os.makedirs(path)
         for proto_file in files:
             file_path = os.path.join(path, proto_file.name)
+            os.makedirs(os.path.split(file_path)[0], exist_ok=True)
             self._save_proto_file(file_path, proto_file.content)
 
     def load_main_proto_files(self) -> None:
         """Loads and compiles the main Exonum proto files."""
         if self._proto_dir is None:
-            raise RuntimeError("Attempt to use unititialized ProtobufLoader")
+            err_msg = "Attempt to use uninitialized ProtobufLoader."
+            logger.critical(err_msg)
+            raise RuntimeError(err_msg)
 
         # This method is not intended to be used by end users, but it is OK to call it here.
         # pylint: disable=protected-access
@@ -187,23 +219,27 @@ class ProtobufLoader:
         proto_dir = os.path.join(self._proto_dir, "exonum_modules", "main")
         self.protoc.compile(main_dir, proto_dir)
 
-    def load_service_proto_files(self, runtime_id: int, service_name: str) -> None:
-        """Loads and compiles proto files for a service.
-
-        `runtime_id` for exonum 0.12 and below should always be 0."""
+    def load_service_proto_files(self, runtime_id: int, artifact_name: str, artifact_version: str) -> None:
+        """Loads and compiles proto files for a service."""
         if self._proto_dir is None:
+            logger.critical("Attempt to use unititialized ProtobufLoader.")
             raise RuntimeError("Attempt to use unititialized ProtobufLoader")
 
         # This method is not intended to be used by end users, but it is OK to call it here.
         # pylint: disable=protected-access
-        proto_contents = self.client.get_proto_sources_for_artifact(runtime_id, service_name)
+        proto_contents = self.client.get_proto_sources_for_artifact(runtime_id, artifact_name, artifact_version)
 
-        # Save proto_sources in proto/service_name directory:
-        service_module_name = re.sub(r"[-. :/]", "_", service_name)
+        # Save proto_sources in proto/artifact_name directory:
+        artifact_name_and_version = f"{artifact_name}:{artifact_version}"
+        service_module_name = re.sub(r"[-. :/]", "_", artifact_name_and_version)
         service_dir = os.path.join(self._proto_dir, "proto", service_module_name)
         self._save_files(service_dir, proto_contents)
 
         # Call protoc to compile proto sources:
         main_dir = os.path.join(self._proto_dir, "proto", "main")
         proto_dir = os.path.join(self._proto_dir, "exonum_modules", service_module_name)
-        self.protoc.compile(service_dir, proto_dir, include=main_dir)
+        if runtime_id != PYTHON_RUNTIME:
+            self.protoc.compile(service_dir, proto_dir, include=main_dir)
+        else:
+            # Python services do not rely on the `includes` from the exonum core.
+            self.protoc.compile(service_dir, proto_dir)
